from email import header
import hmac
import base64
import hashlib
import json
import time
from wsgiref import headers
import requests
from urllib.parse import urlencode
from ttxt.utils import general
from ttxt.base import baseSpotExchange
from ttxt.types import baseTypes


class bitget(baseSpotExchange.BaseSpotExchange):
    def __init__(self, key, secret, password, **kwargs):
        super().__init__(key, secret, **kwargs)
        self.domain_url = "https://api.bitget.com"
        self.password = password
        self.success_sode = '00000'
        
    def _getSymbol(self, symbol):
        return symbol.replace("/", "")
    
    def _getUserSymbol(self, symbol):
        raise NotImplementedError("method not implemented")
    
    ## Auth 
    def sign(self, message):
        mac = hmac.new(bytes(self.secret, encoding='utf-8'), bytes(message, encoding='utf-8'), digestmod='sha256')
        d = mac.digest()
        return base64.b64encode(d)
    
    def pre_hash(self, timestamp, method, request_path, queryString, body=None):
        if not body:
            return str(timestamp) + str.upper(method) + request_path + queryString
        return str(timestamp) + str.upper(method) + request_path + body
    
    def parse_params_to_str(self, params):
        params = [(key, val) for key, val in params.items()]
        params.sort(key=lambda x:x[0])
        url = "?" + urlencode(params)
        if url == '?':
            return ''
        return url
    
    def generate_timestamp(self):
        return int(time.time() * 10**3)
    
    def _signedRequest(self, method, request_path, queryString, body):
        timeMs = self.generate_timestamp()
        if method == "POST":
            body = json.dumps(body)
            queryString = json.dumps(queryString) if queryString is not None else ''
            signature = self.sign(self.pre_hash(timeMs, method, request_path, queryString ,body))
            headers = {"ACCESS-KEY": self.key, "ACCESS-SIGN": signature, "ACCESS-PASSPHRASE": self.password, "ACCESS-TIMESTAMP": str(timeMs), "locale": "en-US", "Content-Type": "application/json"}
            url = self.domain_url+request_path
            try:
                response = requests.post(url, headers=headers, data=body)
                return response.json()
            except Exception as e:
                raise e
        if method == "GET":
            params = queryString
            body = ""
            request_path = request_path + self.parse_params_to_str(params)
            signature = self.sign(self.pre_hash(timeMs, "GET", request_path, str(body)))
            headers = {"ACCESS-KEY": self.key, "ACCESS-SIGN": signature, "ACCESS-PASSPHRASE": self.password, "ACCESS-TIMESTAMP": str(timeMs), "locale": "en-US", "Content-Type": "application/json"}
            url = self.domain_url+request_path
            try:
                response = requests.get(url, headers=headers)
                return response.json()
            except Exception as e:
                raise e
    
    def _unsignedRequest(self, method, apiUrl, params):
        url = self.domain_url + apiUrl
        if method == 'GET':
            try:
                response = requests.request('get', url, params=params)
                return response.json()
            except Exception as e:
                raise e
        else:
            raise Exception(f"{method} Method not supported for unsigned calls")
    
    ## parsers
    def _parseBalance(self, balData):
        parsedBal = {"free": {}, "total": {}}
        if balData['code'] != '00000': 
            raise Exception(balData['msg'])
        data = balData.get("data", None)
        if data is not None:
            for element in data:
                parsedBal["free"][element["coin"]] = element.get("available", None)
                parsedBal["total"][element["coin"]] = str(float(element['available'])+float(element['frozen']))
        return parsedBal
    
    def _parseCreateorder(self, order):
        if order['code'] != '00000': 
            raise Exception(order['msg'])
        parsedOrder = {'id': None, "symbol": None, "amount": None, "side": None, "timestamp": None, "status": None, "orderJson": None}
        if "data" in order and order["data"] != {}:
            parsedOrder['id'] = order['data']['orderId']
            parsedOrder["orderJson"] = json.dumps(order)
        return parsedOrder
    
    def _parseOpenOrders(self, orderjson):
        if orderjson['code'] != '00000': 
            raise Exception(orderjson['msg'])
        parsedOrderList = []    
        if "data" in orderjson and orderjson['data'] != {}:
            for order in orderjson["data"]:
                order['price'] = order['priceAvg']
                parsedOrderList.append(self._parseOrder(order))
        return parsedOrderList
    
    def _parseOrder(self, order):
        parsedOrder = {}
        parsedOrder["id"] = order.get('orderId', None)
        parsedOrder["symbol"] = order.get('symbol', None)
        parsedOrder["price"] = float(order["price"]) if "price" in order else None 
        parsedOrder["amount"] = float(order["size"]) if "size" in order else None
        parsedOrder["side"] = order['side'].lower() if "side" in order else None
        parsedOrder["timestamp"] = int(order["cTime"]) if "cTime" in order else None
        parsedOrder["status"] = order.get("status", None)
        parsedOrder["orderJson"] = json.dumps(order) if order else None 
        return parsedOrder
    
    def _parseFetchOrder(self, data):
        parsedOrder = {"id": None, "symbol": None, "price": None, "amount": None, "takerOrMaker": None, "datetime": None, "fee": None, "fee_currency": None,
                       "side": None, "timestamp": None, "status": None}
        if "data" in data and data["data"] != {}:
            order = data['data']
            if type(order) == list:
                order = order[0]
            parsedOrder["id"] = order["orderId"]
            parsedOrder["tradeId"] = order.get("tradeId", None)
            parsedOrder["symbol"] = order["symbol"]
            if "price" in order:
                parsedOrder["price"] = float(order["price"])
            else:
                parsedOrder["price"] = float(order["priceAvg"])
            parsedOrder['takerOrMaker'] = order.get('tradeScope', None)
            parsedOrder["amount"] = float(order["size"])
            parsedOrder["side"] = order["side"].lower()
            parsedOrder["timestamp"] = int(order["uTime"])
            parsedOrder["datetime"] = general.ts_to_datetime(parsedOrder["timestamp"])
            parsedOrder["status"] = order.get('status', None)
            parsedOrder['fee'] = order.get("feeDetail", None).get('totalFee', None) if order.get("feeDetail", None) is not None else None
            parsedOrder["fee_currency"] = order["feeDetail"].get('feeCoin', None)
        return parsedOrder
    
    def _parseFetchTrades(self, orderjson):
        if orderjson['code'] != '00000': 
            raise Exception(orderjson['msg'])
        parsedTradesList = []
        if "data" in orderjson and orderjson["data"] != []:
            for order in orderjson["data"]:
                parsedTradesList.append(self._parseFetchOrder({'data': order}))
        return parsedTradesList
    
    def _parseCancelorder(self, order):
        if order['code'] != '00000': 
            raise Exception(order['msg'])
        parsedOrder = {'id': None, "symbol": None, "amount": None, "side": None, "timestamp": None, "status": None, "orderJson": None}
        if "data" in order and order["data"] != {}:
            parsedOrder['id'] = order['data']['orderId']
        return parsedOrder
    
    ## Exchange functions 
    def fetch_ticker(self, symbol):
        raise NotImplementedError("method not implemented")
    
    def fetch_balance(self, params={}) -> baseTypes.Balances:
        apiUrl = "/api/v2/spot/account/assets"
        try:
            resp = self._signedRequest('GET', apiUrl, params, '')
            return self._parseBalance(resp)
        except Exception as e:
            raise e
        
    def fetch_my_trades(self, symbol=None, since=None, limit=None, params={}): 
        apiUrl = "/api/v2/spot/trade/fills"
        try:
            params = {
                "symbol": self._getSymbol(symbol)
            }
            if since:
                params["startTime"] = since
            if 'endTime' in params and params['endTime']:
                params['endTime'] = params['endTime']
            if limit:
                params['limit'] = limit
            response = self._signedRequest('GET', request_path=apiUrl, queryString=params, body='')
            return self._parseFetchTrades(response)
        except Exception as e:
            raise e
    
    def create_order(self, symbol, side, amount, order_type, price=None, params={}): 
        apiUrl = "/api/v2/spot/trade/place-order"
        params = {
            "symbol": self._getSymbol(symbol),
            "force": "GTC",
            "side": side.lower(),
            "size": str(amount)
        }
        try:
            if order_type == 'limit':
                params["orderType"] = "limit"
                params['price'] = str(price)
            elif order_type == 'market':
                params["orderType"] = "market"
            params.update(params) 
            response = self._signedRequest('POST', request_path=apiUrl, body=params, queryString='')
            return self._parseCreateorder(response)
        except Exception as e:
            raise e
    
    def create_market_order(self, symbol, side, amount, params={}):
        apiUrl = "/api/v2/spot/trade/place-order"
        params = {
            "symbol": self._getSymbol(symbol),
            "orderType": "market",
            "side": side.lower(),
            "size": str(amount)
        }
        try:
            params.update(params) 
            response = self._signedRequest('POST', request_path=apiUrl, body=params, queryString='')
            return self._parseCreateorder(response)
        except Exception as e:
            raise e
    
    def create_limit_order(self, symbol, side, amount, price, params={}):
        apiUrl = "/api/v2/spot/trade/place-order"
        params = {
            "symbol": self._getSymbol(symbol),
            "orderType": "limit",
            "force": "GTC",
            "side": side.lower(),
            "price": str(price),
            "size": str(amount)
        }
        try:
            params.update(params) 
            response = self._signedRequest('POST', request_path=apiUrl, body=params, queryString='')
            return self._parseCreateorder(response)
        except Exception as e:
            raise e
    
    def cancel_order(self, id=None, symbol=None, params={}):
        apiUrl = "/api/v2/spot/trade/cancel-order"
        ticker = self._getSymbol(symbol)
        try:
            params={}
            if id is not None:
                params['orderId'] = id
            if symbol is not None:
                params['symbol'] = str(ticker)
            params.update(params)
            response = self._signedRequest('POST', request_path=apiUrl, queryString='',body=params)
            return self._parseCancelorder(response)
        except Exception as e:
            raise e
    
    def fetch_open_orders(self, symbol=None, kwargs=None):
        apiUrl = "/api/v2/spot/trade/unfilled-orders"
        ticker = self._getSymbol(symbol)
        try:
            params = {
                'symbol': ticker,
                'endTime': self.generate_timestamp(),
                'limit': 100,
            }
            response = self._signedRequest('GET', request_path=apiUrl, queryString=params, body='')
            return self._parseOpenOrders(response)
        except Exception as e:
            raise e
    
    def fetch_order(self, id, symbol = None, params={}):
        apiUrl = "/api/v2/spot/trade/orderInfo"
        try:
            params = {
                'orderId': id
            }
            params.update(params)
            response = self._signedRequest('GET', request_path=apiUrl, queryString=params, body='')
            return self._parseFetchOrder(response)
        except Exception as e:
            raise e

    ## use case specifci request    
    def sub_account_transfer(self, amount, fromUserId, toUserId, coin, fromType="spot", toType="spot"):
        apiUrl = "/api/v2/spot/wallet/subaccount-transfer"
        params = {
            "fromType": fromType,
            "toType": toType,
            "amount": amount,
            "coin": coin,
            "fromUserId": fromUserId,
            "toUserId": toUserId
        }
        try:
            response = self._signedRequest('POST', request_path=apiUrl, body=params, queryString='')
            return response
        except Exception as e:
            raise e


import hmac
import base64
import hashlib
import json
import time
import requests
from ttxt.base import baseSpotExchange
from ttxt.types import baseTypes
from ttxt.utils import general


class cryptocom(baseSpotExchange.BaseSpotExchange):
    def __init__(self, key, secret, **kwargs):
        super().__init__(key, secret, **kwargs)
        self.domain_url = "https://api.crypto.com/v2/" #https://api.crypto.com/exchange/v1/"
        self.MAX_LEVEL = 3

    def _getSymbol(self, symbol):
        return symbol.replace("/", "_")
    
    def _getUserSymbol(self, symbol):
        return symbol.replace("_", "/")

    ## Auth 
    def params_to_str(self, obj, level):
        if level >= self.MAX_LEVEL:
            return str(obj)

        return_str = ""
        for key in sorted(obj):
            return_str += key
            if obj[key] is None:
                return_str += 'null'
            elif isinstance(obj[key], list):
                for subObj in obj[key]:
                    return_str += self.params_to_str(subObj, level + 1)
            else:
                return_str += str(obj[key])
        return return_str
    
    def gen_sign(self, payload_str):
        return hmac.new(bytes(str(self.secret), 'utf-8'), msg=bytes(payload_str, 'utf-8'), digestmod=hashlib.sha256).hexdigest()

    def request_builder(self, params, payload_str, request_path, id, nonce):
        return json.dumps({
            "id": id,
            "method": request_path,
            "api_key": self.key,
            "params": params,
            "nonce": nonce,
            "sig": self.gen_sign(payload_str)
        })

    # def req_param_builder(self, param_str):


    ## Requests 
    def _signedRequest(self, request_path, method, params):
        common_headers = {'Content-Type': 'application/json'}
        if method == "POST":
            try:
                reqId, nonce = general.get_random_int(), general.get_curr_timestamp_ms()
                if params != "" and len(params):
                    param_str = self.params_to_str(params,0)
                    payload_str = request_path + str(reqId) + self.key + param_str + str(nonce)
                    req = self.request_builder(params, payload_str, request_path, reqId, nonce)
                    url = self.domain_url + request_path
                    response = requests.post(url, data=req, headers=common_headers)
                    return response.json()
                else:
                    payload_str = request_path + str(reqId) + self.key + str(nonce)
                    url = self.domain_url + request_path
                    req = self.request_builder(params, payload_str, request_path, reqId, nonce)
                    response = requests.post(url, data=req, headers=common_headers)
                    return response.json()
            except Exception as e:
                raise e
    
    
    def _unsignedRequest(self, method, apiUrl, params):
        url = self.domain_url + apiUrl
        if method == 'GET':
            try:
                response = requests.request('get', url, params=params)
                return response.json()
            except Exception as e:
                raise e
        else:
            raise Exception(f"{method} Method not supported for unsigned calls")

    ## parsers
    def _parseBalance(self, balData):
        parsedBal = {"free": {}, "total": {}}
        if balData['code'] != 0: raise Exception(balData['message'])
        for bal in balData['result']['accounts']:
            parsedBal["free"][bal["currency"]] = float(bal["available"])
            parsedBal["total"][bal["currency"]] = float(bal['balance'])
        return parsedBal
    
    def _parseCreateOrder(self, order):
        if order['code'] != 0: raise Exception(order['message'])
        parsedOrder = {}
        parsedOrder["id"] = order['result']["order_id"]
        parsedOrder["symbol"] = None
        parsedOrder["price"] = None
        parsedOrder["amount"] = None
        parsedOrder["side"] = None
        parsedOrder["timestamp"] = None
        parsedOrder["status"] = None
        parsedOrder["orderJson"] = json.dumps(order)
        return parsedOrder
    
    def _parseCancelOrder(self, order):
        if order['code'] != 0: raise Exception(order['message'])
        parsedOrder = {}
        parsedOrder["id"] = None
        parsedOrder["symbol"] = None
        parsedOrder["price"] = None
        parsedOrder["amount"] = None
        parsedOrder["side"] = None
        parsedOrder["timestamp"] = None
        parsedOrder["status"] = None
        parsedOrder["orderJson"] = json.dumps(order)
        return parsedOrder
    
    def _parseOrderbook(self, orderbookData):
        parsedData = {}
        parsedData["datetime"] = None
        parsedData["timestamp"] = orderbookData.get("current", None)
        parsedData["nonce"] = orderbookData.get("id", None)
        if "bids" in orderbookData:
            parsedData["bids"] = [[float(d[0]), float(d[1])] for d in orderbookData["bids"]]
        else: parsedData["bids"] = []
        if "asks" in orderbookData:
            parsedData["asks"] = [[float(d[0]), float(d[1])] for d in orderbookData["asks"]]
        else: parsedData["asks"] = []
        return parsedData
    
    def _parseOrder(self, order):
        parsedOrder = {}
        parsedOrder["id"] = order.get('order_id', None)
        parsedOrder["symbol"] = self._getUserSymbol(order["instrument_name"]) if "instrument_name" in order else None
        parsedOrder["price"] = float(order["price"]) if "price" in order else None 
        parsedOrder["amount"] = float(order["quantity"]) if "quantity" in order else None
        parsedOrder["side"] = order['side'].lower()
        parsedOrder["timestamp"] = int(order["create_time"]) if "create_time" in order else None
        parsedOrder["status"] = order.get("status", None)
        parsedOrder["orderJson"] = json.dumps(order) if order else None 
        return parsedOrder
    
    def _parseOpenOrders(self, orders):
        if orders['code'] != 0: raise Exception(orders['message'])
        parsedOrderList = []
        for order in orders['result']['order_list']:
            parsedOrderList.append(self._parseOrder(order))
        return parsedOrderList
    
    def _parseTrades(self, trades):
        def parseTrade(trade):
            parsedTrade = {}
            parsedTrade["id"] = trade.get("order_id", None)
            parsedTrade["tradeId"] = trade.get("trade_id", None)
            parsedTrade["side"] = trade["side"].lower()
            parsedTrade["price"] = trade.get("traded_price", None)
            parsedTrade["amount"] = trade.get("traded_quantity", None)
            parsedTrade["takerOrMaker"] = None # cc doesnt return taker or maker role
            if "create_time" in trade:
                parsedTrade["timestamp"] = int(float(trade["create_time"]))
                parsedTrade["datetime"] = general.ts_to_datetime(parsedTrade["timestamp"])
            else: 
                parsedTrade["timestamp"] = None
                parsedTrade["datetime"] = None
            parsedTrade["fee"] = trade.get("fee", None)
            parsedTrade["fee_currency"] = trade.get("fee_currency", None)
            return parsedTrade
        if trades['code'] != 0: raise Exception(trades['message'])
        parsedTradeList = []
        for trade in trades['result']['trade_list']:
            parsedTradeList.append(parseTrade(trade))
        return parsedTradeList

    ## Exchange functions 

    def create_order(self, symbol, type, side, amount, price=None, params={}):
        try:
            if type == "market":
                self.create_market_order(symbol, amount, side, params={})
            elif type == "limit":
                self.create_limit_order(symbol, amount, side, price, params={})
        except Exception as e:
            raise e
        
    def create_limit_order(self, symbol, amount, side, price, params={}):
        body = {
            "instrument_name": self._getSymbol(symbol),
            "side": side.upper(),
            "type": "LIMIT",
            "price": price,
            "quantity": amount
        }
        body.update(params)
        apiUrl = "private/create-order"
        try:
            resp = self._signedRequest(method='POST', request_path=apiUrl, params=body)
            return self._parseCreateOrder(resp)
        except Exception as e:
            raise e

    def create_market_order(self, symbol, amount, side, params={}):
        if side == 'buy':
            body = {
                "instrument_name": self._getSymbol(symbol),
                "side": side.upper(),
                "type": "MARKET",
                "notional": amount
            }
        else:
            body = {
                "instrument_name": self._getSymbol(symbol),
                "side": side.upper(),
                "type": "MARKET",
                "quantity": amount
            }
        body.update(params)
        apiUrl = "private/create-order"
        try:
            resp = self._signedRequest(method='POST', request_path=apiUrl, params=body)
            return self._parseCreateOrder(resp)
        except Exception as e:
            raise e
    
    def fetch_order(self, id, symbol=None):
        raise NotImplementedError("method not implemented")

    def fetch_open_orders(self, symbol=None,  kwargs=None):
        apiUrl = "private/get-open-orders"
        params = {"instrument_name": self._getSymbol(symbol)}
        try:
            resp = self._signedRequest(method='POST', request_path=apiUrl, params=params)
            return self._parseOpenOrders(resp)
        except Exception as e:
            raise e
        
    def cancel_order(self, id, symbol=None):
        apiUrl = "private/cancel-order"
        params = {"instrument_name": self._getSymbol(symbol), "order_id": id}
        try:
            resp = self._signedRequest(method='POST', request_path=apiUrl, params=params)
            return self._parseCancelOrder(resp)
        except Exception as e:
            raise e

    def fetch_ticker(self, symbol: str, params={}) -> baseTypes.Ticker:
        raise NotImplementedError("method not implemented")
    
    def fetch_balance(self, params={}) -> baseTypes.Balances:
        apiUrl = "private/get-account-summary"
        try:
            resp = self._signedRequest(method='POST', request_path=apiUrl, params={})
            return self._parseBalance(resp)
        except Exception as e:
            raise e
        
    def fetch_my_trades(self, symbol=None, since=None, limit=None, params={}):
        if symbol:
            params['instrument_name'] = self._getSymbol(symbol)
        if limit:
            params['page_size'] = limit
        if since:
            params['start_ts'] = int(since)
        apiUrl = "private/get-trades"
        try:
            resp = self._signedRequest(method='POST', request_path=apiUrl, params={})
            return self._parseTrades(resp)
        except Exception as e:
            raise e
        
    def fetch_order_book(self, symbol: str, limit=None, params={}) -> baseTypes.OrderBook:
        apiUrl = f"/spot/order_book"
        queryParams = {"currency_pair": self._getSymbol(symbol)}
        if limit is not None:
            queryParams["limit"] = int(limit)
        queryParams.update(params)
        try:
            resp = self._unsignedRequest('GET', apiUrl, params=queryParams)
            return self._parseOrderbook(resp)
        except Exception as e:
            raise e
    
    # parsed OHLCV = [[ts_ms, o, h ,l ,c, v],[]]
    def fetch_ohlcv(self, symbol: str, timeframe='1m', since = None, limit = None, params={}):
        raise NotImplementedError("method not implemented")


class BaseFuturesExchange:
    def __init__(self, key, secret, **kwargs):
        self.key = key
        self.secret = secret
        self.additional_params = kwargs

    def fetch_ticker(self, symbol):
        raise NotImplementedError("Subclasses must implement this method")
    
    def fetch_ohlcv(self, symbol: str, timeframe='1m', since = None, limit = None, params={}):
        raise NotImplementedError("Subclasses must implement this method")
    
    def fetch_balance(self, symbol):
        raise NotImplementedError("Subclasses must implement this method")

    def create_order(self, symbol, quantity, price, order_type):
        raise NotImplementedError("Subclasses must implement this method")
    
    def fetch_order(self, id: str, symbol = None, params={}):
        raise NotImplementedError("Subclasses must implement this method")
    
    def fetch_open_orders(self, symbol=None,  kwargs=None):
        raise NotImplementedError("Subclasses must implement this method")
    
    def cancel_order(self, id: str):
        raise NotImplementedError("Subclasses must implement this method")
    
    def set_leverage(self, leverage, symbol, params={}):
        raise NotImplementedError("Subclasses must implement this method")
    
    def fetch_order_book(self, symbol: str, limit=None, params={}):
        raise NotImplementedError("Subclasses must implement this method")


import hmac
import base64
import hashlib
import json
import time
import requests
from datetime import datetime as dt
from ttxt.base import baseSpotExchange
from ttxt.types import baseTypes
from ttxt.utils import general

'''
kwards = {
    "category": "",
    "recv_window": ""
}
'''
class bingx(baseSpotExchange.BaseSpotExchange):
    def __init__(self, key, secret, **kwargs):
        super().__init__(key, secret, **kwargs)
        self.domain_url = "https://open-api.bingx.com"
        self.prefix = "/api/v4"
        self.category = kwargs["category"] if "category" in kwargs else "linear"
        self.recv_window = kwargs["recv_window"] if "recv_window" in kwargs else "5000"
        self.account_type = kwargs["account_type"] if "account_type" in kwargs else "UNIFIED"
        self.max_retries = 5
        self.symbolLotSize = {}

    def _getSymbol(self, symbol):
        return symbol.replace("/", "-")
    
    def _getUserSymbol(self, symbol):
        return symbol.replace("-", "/")
    
    ## Auth 
    def get_expire(self):
        return int((time.time() + 1) * 1000)  # websockets use seconds
    
    def generate_timestamp(self):
        """
        Return a millisecond integer timestamp.
        """
        return int(time.time() * 10**3)
    
    def _get_sign(self, payload):
        signature = hmac.new(self.secret.encode("utf-8"), payload.encode("utf-8"), digestmod=hashlib.sha256).hexdigest()
        return signature
    
    def _parseParams(self, paramsMap):
        sortedKeys = sorted(paramsMap)
        paramsStr = "&".join(["%s=%s" % (x, paramsMap[x]) for x in sortedKeys])
        return paramsStr+"&timestamp="+str(int(time.time() * 1000))

    def _send_request(self, method, path, params, payload):
        params = self._parseParams(params)
        url = "%s%s?%s&signature=%s" % (self.domain_url, path, params, self._get_sign(params))
        headers = {
            'X-BX-APIKEY': self.key,
        }
        try:
            response = requests.request(method, url, headers=headers, data=payload)
            return response.json()
        except Exception as e:
            raise e
        
    def _prepare_params(self, params):
        payload = "&".join(
                [
                    str(k) + "=" + str(v)
                    for k, v in sorted(params.items())
                    if v is not None
                ]
            )
        return payload

    def _unsignedRequest(self, method=None, path=None, query=None, auth=False):
        path = self.domain_url + path
        if query is None:
            query = {}
        # Store original recv_window.
        recv_window = self.recv_window
        # Bug fix: change floating whole numbers to integers to prevent
        # auth signature errors.
        if query is not None:
            for i in query.keys():
                if isinstance(query[i], float) and query[i] == int(query[i]):
                    query[i] = int(query[i])
        # Send request and return headers with body. Retry if failed.
        retries_attempted = self.max_retries
        req_params = None
        while True:
            retries_attempted -= 1
            if retries_attempted < 0:
                raise Exception(
                    "Bad Request. Retries exceeded maximum."
                )
            req_params = self._prepare_params(query)
            # Authenticate if we are using a private endpoint.
            headers = {}
            if method == "GET":
                try:
                    if req_params:
                        client = requests.Session()
                        r = client.prepare_request(requests.Request(method, path, headers=headers))
                        #esp = client.send(r, timeout=60) 
                        resp = requests.get(path + f"?{req_params}", headers=headers)
                    else:
                        r = requests.get(path, headers=headers)
                    return resp.json()
                except Exception as e:
                    raise e
            if method == "POST":
                r = requests.post( path, data=req_params, headers=headers)
                return r.json()
            if method == "DELETE":
                r = requests.delete( path, data=req_params, headers=headers)
                return r.json()

    ## parsers
    def _parseBalance(self, balData):
        if balData['code'] != 0: raise Exception(balData['msg'])
        parsedBal = {"free": {}, "total": {}}
        for balDataEle in balData['data']['balances']:
            parsedBal["free"][balDataEle["asset"]] = float(balDataEle["free"])
            parsedBal["total"][balDataEle["asset"]] = float(balDataEle["locked"]) + parsedBal["free"][balDataEle["asset"]]
        return parsedBal

    def _parseOrder(self, order):
        parsedOrder = {"id": None, "symbol": None, "price": None, "amount": None, 
                       "side": None, "timestamp": None, "status": None, "orderJson": None}
        parsedOrder["id"] = order["orderId"]
        parsedOrder["symbol"] = self._getUserSymbol(order["symbol"])
        parsedOrder["price"] = order["price"]
        parsedOrder["amount"] = order["origQty"]
        parsedOrder["side"] = order["side"].lower()
        parsedOrder["timestamp"] = order["time"]
        parsedOrder["status"] = order["status"]
        parsedOrder["orderJson"] = json.dumps(order)
        return parsedOrder
    
    '''
    {'code': 0, 'msg': '', 'debugMsg': '', 'data': {'symbol': 'GALA-USDT', 'orderId': 1765446279820824576, 'price': '0.04', 'stopPrice': '0', 
    'origQty': '100', 'executedQty': '0', 'cummulativeQuoteQty': '0', 'status': 'CANCELED', 'type': 'LIMIT', 'side': 'BUY'}}
    '''
    def _parseCancelOrder(self, order):
        if order['code'] != 0: raise Exception(order['msg'])
        order = order['data']
        parsedOrder = {"id": None, "symbol": None, "price": None, "amount": None, 
                       "side": None, "timestamp": None, "status": None, "orderJson": None}
        parsedOrder["id"] = order["orderId"]
        parsedOrder["symbol"] = self._getUserSymbol(order["symbol"])
        parsedOrder["price"] = order["price"]
        parsedOrder["amount"] = order["origQty"]
        parsedOrder["side"] = order["side"].lower()
        parsedOrder["status"] = order["status"]
        parsedOrder["orderJson"] = json.dumps(order)
        return parsedOrder
    
    def _parseCreateOrder(self, order):
        if order['code'] != 0: raise Exception(order['msg'])
        parsedOrder = {}
        order = order["data"]
        parsedOrder["id"] = order["orderId"]
        parsedOrder["symbol"] = self._getUserSymbol(order["symbol"])
        parsedOrder["price"] = order["price"]
        parsedOrder["amount"] = order["origQty"]
        parsedOrder["side"] = order["side"].lower()
        parsedOrder["timestamp"] = int(order["transactTime"])
        parsedOrder["status"] = order["status"]
        parsedOrder["orderJson"] = json.dumps(order)
        return parsedOrder

    def _parseOpenOrders(self, orders):
        if orders['code'] != 0: raise Exception(orders['msg'])
        parsedOrderList = []
        for order in orders["data"]["orders"]:
            parsedOrderList.append(self._parseOrder(order))
        return parsedOrderList
    
    def _parseTrades(self, trades):
        def parseTrade(trade):
            parsedTrade = {}
            parsedTrade["symbol"] = self._getUserSymbol(trade["symbol"])
            parsedTrade["id"] = trade.get("orderId", None)
            parsedTrade["tradeId"] = trade.get("id", None)
            parsedTrade["side"] = "buy" if trade["isBuyer"] else "sell"
            parsedTrade["price"] = trade.get("price", None)
            parsedTrade["amount"] = trade.get("qty", None)
            parsedTrade["takerOrMaker"] = "maker" if trade["isMaker"] else "taker"
            if "time" in trade:
                parsedTrade["timestamp"] = trade["time"]
                parsedTrade["datetime"] = general.ts_to_datetime(parsedTrade["timestamp"])
            else: 
                parsedTrade["timestamp"] = None
                parsedTrade["datetime"] = None
            parsedTrade["fee"] = trade.get("commission", None)
            parsedTrade["fee_currency"] = trade.get("commissionAsset", None)
            return parsedTrade
        if trades['code'] != 0: raise Exception(trades['msg'])
        parsedTradeList = []
        for trade in trades['data']['fills']:
            parsedTradeList.append(parseTrade(trade))
        return parsedTradeList


    def _parseFetchedOrder(self, order):
        pass

    ## Exchange functions 
    def create_order(self, symbol, type, side, amount, price=None, params={}):
        try:
            if type == "market":
                self.create_market_order(symbol, amount, side, params={})
            elif type == "limit":
                self.create_limit_order(symbol, amount, side, price, params={})
        except Exception as e:
            raise e
    
    def create_limit_order(self, symbol, amount, side, price, params={}):
        ticker = self._getSymbol(symbol)
        params = {
            "symbol": ticker,
            "side": side.upper(),
            "type": "LIMIT",
            "price": float(price),
            "quantity": float(amount),
            "timestamp": self.generate_timestamp()
        }
        params.update(params) 
        apiUrl = "/openApi/spot/v1/trade/order"
        response = self._send_request(method='POST', path=apiUrl, params=params, payload={})
        return self._parseCreateOrder(response)
    
    def create_market_order(self, symbol, amount, side, params={}):
        ticker = self._getSymbol(symbol)
        params = {
            "symbol": ticker,
            "side": side.upper(),
            "type": "MARKET",
            "quantity": float(amount),
            "timestamp": self.generate_timestamp()
        }
        params.update(params) 
        apiUrl = "/openApi/spot/v1/trade/order"
        response = self._send_request(method='POST', path=apiUrl, params=params, payload={})
        return self._parseCreateOrder(response)
    
    def fetch_order(self, id, symbol=None):
        apiUrl = "/openApi/spot/v1/trade/query"
        queryParams = {"orderId": id, "symbol": symbol}  # maybe have a set symbol function ?
        try:
            resp = resp = self._send_request(method='GET', path=apiUrl, params=queryParams, payload=None)
            return resp #self._parseOrder(resp)
        except Exception as e:
            raise e

    def fetch_open_orders(self, symbol=None,  kwargs=None):
        apiUrl = "/openApi/spot/v1/trade/openOrders"
        queryParams = {"timestamp": self.generate_timestamp()}
        try:
            resp = self._send_request(method='GET', path=apiUrl, params=queryParams, payload=None)
            return self._parseOpenOrders(resp)
        except Exception as e:
            raise e
        
    def cancel_order(self, id, symbol=None, params={}):
        apiUrl = "/openApi/spot/v1/trade/cancel"
        queryParams = {"orderId": id, "symbol": self._getSymbol(symbol)}
        try:
            resp = self._send_request(method='GET', path=apiUrl, params=queryParams, payload=None)
            return self._parseCancelOrder(resp)
        except Exception as e:
            raise e

    def fetch_ticker(self, symbol: str, params={}) -> baseTypes.Ticker:
        apiUrl = "/openApi/spot/v1/common/symbols"
        queryParams = {"symbol": self._getSymbol(symbol)}
        try:
            resp = self._unsignedRequest(method='GET', path=apiUrl, query=queryParams)
            return resp # parse this response into Ticker
        except Exception as e:
            raise e
    
    # params = {"startTime": 1702512246000, "endTime": 1702512248000, "limit": 100}
    def fetch_ohlcv(self, symbol, interval, params={}):
        apiUrl = "/openApi/spot/v2/market/kline"
        queryParams = {"symbol": self._getSymbol(symbol), "interval": interval}
        try:
            resp = self._unsignedRequest(method='GET', path=apiUrl, query=queryParams)
            return resp # parse this response into Ticker
        except Exception as e:
            raise e
    
    def fetch_balance(self, params={}) -> baseTypes.Balances:
        apiUrl = '/openApi/spot/v1/account/balance'
        queryParams = {}
        try:
            resp = self._send_request(method='GET', path=apiUrl, params=queryParams, payload=None)
            return self._parseBalance(resp)
        except Exception as e:
            raise e
        
    def fetch_my_trades(self, symbol=None, since=None, limit=None, params={}):
        apiUrl = '/openApi/spot/v1/trade/myTrades'
        queryParams = {
            "symbol": self._getSymbol(symbol),
        }
        if since:
            queryParams["startTime"] = since
        if 'endTime' in params and params['endTime']:
                params['endTime'] = params['endTime']
        if limit:
            params['limit'] = limit
        try:
            resp = self._send_request(method='GET', path=apiUrl, params=queryParams, payload=None)
            return self._parseTrades(resp)
        except Exception as e:
            raise e


from setuptools import setup, find_packages
from os import path
import json
import sys

here = path.abspath(path.dirname(__file__))
root = path.dirname(here)

setup(
    name='ttxt',
    version='0.0.5.8',
    description='TT exchange interfacing library',
    author='sus',
    packages=find_packages(),
)